#!/usr/bin/env python
# Copyright 2019 XOE Labs (<https://xoe.solutions>)
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points

CONTEXT_SETTINGS = dict(auto_envvar_prefix="ODOOUP")


@with_plugins(iter_entry_points("core_package.cli_plugins"))
@click.group(context_settings=CONTEXT_SETTINGS)
def main():
    """Commandline interface for odooup commands."""
    pass



# Copyright (c) 2015-2018 ACSONE SA/NV
# Copyright (c) 2019 XOE Corp. SAS
# License AGPLv3 (http://www.gnu.org/licenses/agpl-3.0-standalone.html)

import ast
import os

import networkx as nx

from ._helpers import call_cmd

MANIFEST_NAMES = ("__manifest__.py", "__openerp__.py", "__terp__.py")
SKIP_PATHS = ["point_of_sale/tools"]


def _parse_manifest_from_git(repo_path, manifest_object):
    s = call_cmd(
        "git cat-file -p {manifest_object}".format(**locals()),
        echo_cmd=False,
        exit_on_error=False,
        cwd=repo_path,
    )
    return ast.literal_eval(s)


def _find_addons(dir):
    """ yield (addon_dir, addon_name, manifest) """
    project = call_cmd(
        "git ls-tree -r HEAD", echo_cmd=False, exit_on_error=False, cwd=dir
    ).split("\n")
    project = [i.split() for i in project]
    manifests = {
        (".", i[2], i[3])
        for i in project
        if i[1] == "blob"
        and any(M in i[3] for M in MANIFEST_NAMES)
        and not any(S in i[3] for S in SKIP_PATHS)
    }
    project_submodule_paths = [i[3] for i in project if i[1] == "commit"]
    for path in project_submodule_paths:
        submodule = call_cmd(
            "git ls-tree -r HEAD", echo_cmd=False, exit_on_error=False, cwd=path
        ).split("\n")
        submodule = [i.split() for i in submodule]
        manifests |= {
            (path, i[2], os.path.join(path, i[3]))
            for i in submodule
            if i[1] == "blob" and any(M in i[3] for M in MANIFEST_NAMES)
        }
    for repo_path, manifest_object, manifest_path in manifests:
        module_path = os.path.dirname(manifest_path)
        manifest = _parse_manifest_from_git(repo_path, manifest_object)
        yield os.path.dirname(module_path), os.path.basename(module_path), manifest


def get_graph(dir):
    g = nx.DiGraph()
    for namespace, name, manifest in _find_addons(dir):
        g.add_node(name, manifest=manifest, namespace=namespace)
        edges = zip(
            manifest.get("depends", []), [name] * len(manifest.get("depends", []))
        )
        g.add_edges_from(edges)
    return g



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
from collections import OrderedDict

import click

from ._helpers import call_cmd, replace_in_file
from ._installers import install_tools
from .clone import clone_submodule_to_target, get_vendor_target

ODOO_VERSIONS = OrderedDict(
    [("10", "10.0"), ("11", "11.0"), ("12", "12.0"), ("m", "master")]
)

IS_GIT_URL_REGEXP = (
    r"(?:git|ssh|https?|git@[-\w.]+):(\/\/)?(.*?)(\.git)(\/?|\#[-\d\w._]+?)$"
)


class OdooVersionChoice(click.types.Choice):
    name = "odoo-version"

    def __init__(self, versions, case_sensitive=True):
        self.versions = versions
        super(OdooVersionChoice, self).__init__(self.versions.keys(), case_sensitive)

    def convert(self, value, param, ctx):
        value = super(OdooVersionChoice, self).convert(value, param, ctx)
        return self.versions[value]


class GitRepo(click.types.StringParamType):
    name = "git-repo"

    def convert(self, value, param, ctx):
        value = super(GitRepo, self).convert(value, param, ctx)
        found = re.match(IS_GIT_URL_REGEXP, value)

        if not found:
            self.fail(click.style(value, fg="red") + " is not a git url", param, ctx)

        return value


def ask_for_additional_repos():
    additional_repos = []
    fist_repo = True
    while True:
        msg = "Add additional repo" if fist_repo else "Add another repo"
        add_new = click.confirm(msg, default=False)
        if not add_new:
            break
        repo = click.prompt("Enter Repo", type=GitRepo())
        fist_repo = False
        additional_repos += [repo]
    return additional_repos


@click.command()
@click.option(
    "--odoo-version",
    type=OdooVersionChoice(ODOO_VERSIONS),
    prompt="Please select Odoo version: \n Available: "
    + ", ".join(["{}".format(v) for v in ODOO_VERSIONS.values()])
    + "\n Select:",
)
@click.option(
    "--is-enterprise",
    is_flag=True,
    default=False,
    prompt="Use Odoo Enterprise (access required)",
    help="Clone enterprise repository (access required).",
)
@click.argument("project", required=True)
def init(odoo_version, is_enterprise, project):
    """ Bootstrap you Odoo project """
    additional_repos = ask_for_additional_repos()
    project = project.lower()  # docker doesn't permit uppercase project names

    click.echo("")
    click.secho("Your choices: ", fg="black", bg="bright_green", bold=True)
    click.echo("")
    click.echo("Odoo Edition:      ", nl=False)
    ee = "EE (Enterprise Edition)"
    ce = "CE (Community Edition)"
    click.secho(
        "Odoo " + odoo_version + " " + (ee if is_enterprise else ce),
        fg="yellow",
        bold=True,
    )
    click.echo("")
    click.secho("Get some work done ...", fg="black", bg="bright_green", bold=True)
    click.echo("")

    install_tools()

    if (
        project
        and call_cmd("git rev-parse --is-inside-work-tree", exit_on_error=False)
        == "true"
    ):
        click.get_current_context().fail(
            "You are trying to launch a new project inside a git repository"
        )
    elif project:
        call_cmd(
            "git clone https://github.com/xoe-labs/dockery-odoo-scaffold.git {}".format(
                project
            )
        )
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, project)
        click.echo("Switching to: " + click.style(final_directory, fg="yellow"))
        os.chdir(final_directory)
        call_cmd("git remote rename origin scaffold")
        call_cmd("git branch --unset-upstream")

    # Repo cloning
    for repo_url in additional_repos:
        clone_submodule_to_target(odoo_version, repo_url, get_vendor_target(repo_url))
    clone_submodule_to_target(
        odoo_version, "https://github.com/odoo/odoo.git", "./vendor/odoo/cc"
    )
    if is_enterprise:
        clone_submodule_to_target(
            odoo_version, "https://github.com/odoo/enterprise.git", "./vendor/odoo/ee"
        )

    # Seed Placeholders
    replacements = {
        ("Dockerfile", ".env"): [
            {"from": "{{ PROJECT }}", "to": project},
            {"from": "{{ DEFAULT_BRANCH }}", "to": odoo_version},
        ]
    }
    for files, rules in replacements.items():
        for rule in rules:
            replace_in_file(files, rule["from"], rule["to"])

    # Git commit
    call_cmd("git add .")
    call_cmd('git commit -m "Customize Project"')

    call_cmd("git config commit.template $(pwd)/.git-commit-template")
    call_cmd("pre-commit install --hook-type pre-commit")
    call_cmd("pre-commit install --hook-type commit-msg")
    call_cmd("pre-commit install --install-hooks")
    click.echo(
        "Next, run: " + click.style("`cd {} && make info`".format(project), fg="yellow")
    )


if __name__ == "__main__":
    init()



import os

import click
import networkx
from networkx import ancestors, dag_longest_path, dag_longest_path_length, subgraph

from ._helpers import call_cmd, get_fs_target
from ._modulegraph import get_graph

DAG_PATH_LENGTH_WARNING_THRESHOLD = 5
DOCKERIGNORE_PLACEHOLDER = "# Autogenerated file content from here ... DO NOT MODIFY"


def _get_sparse_persistence_file(ns):
    return os.path.abspath(os.path.join(ns, "..", ".sparse-" + os.path.basename(ns)))


def _get_ns_from_sparse_persistence_file(ns_path):
    # Get from vendor downstream
    path = "".join(ns_path.partition("vendor")[1:])
    return os.path.join(os.path.dirname(path), os.path.basename(path)[8:])


def _get_sparse_file(ns):
    git_path = call_cmd(
        "git rev-parse --git-dir", echo_cmd=False, exit_on_error=False, cwd=ns
    )
    return os.path.join(git_path, "info", "sparse-checkout")


def _symlink_sparse_file(ns):
    ns_path = _get_sparse_persistence_file(ns)
    sparse_file = _get_sparse_file(ns)
    call_cmd(
        "ln -s {ns_path} {sparse_file}".format(**locals()), exit_on_error=False, cwd=ns
    )


def _warn_path_length(g, deps):
    sub = subgraph(g, deps)
    if dag_longest_path_length(sub) > DAG_PATH_LENGTH_WARNING_THRESHOLD:
        click.secho(
            "The dependency graph of this module is particularily "
            "long (>" + str(DAG_PATH_LENGTH_WARNING_THRESHOLD) + "), "
            "consider refactoring.\nLongest path: ",
            fg="white",
            bg="bright_red",
            bold=True,
        )
        click.secho(" > ".join(dag_longest_path(sub)), fg="white", bold=True)


def ensure_sparse_checkouts(rootpath):
    for root, _, files in os.walk(rootpath):
        for file in files:
            if file.startswith(".sparse-"):
                ns = os.path.join(root, file[8:])
                _symlink_sparse_file(ns)
                call_cmd(
                    "git config core.sparseCheckout True", exit_on_error=False, cwd=ns
                )
                call_cmd("rm -rf *", exit_on_error=False, cwd=ns)
                call_cmd("git reset --hard HEAD", exit_on_error=False, cwd=ns)


def _get_all_sparse_files(g):
    sparse_files = set()
    for module in g:
        node = g.node[module]
        if not node:
            continue
        ns = node["namespace"]
        ns_path = _get_sparse_persistence_file(ns)
        if os.path.isfile(ns_path):
            sparse_files |= {ns_path}
    return sorted(sparse_files)


def _warn_missing_dependencies(g, rootpath):
    for module in g:
        node = g.node[module]
        if not node:
            click.secho(
                "DEPENDENCY INFO: The dependency '{}' was found nowhere "
                "under {}.".format(module, rootpath),
                fg="yellow",
                bold=True,
            )


def _reconcile_auto_install(g):
    all_sparse_files = _get_all_sparse_files(g)
    all_white_listed = []
    auto_install = []
    for module in g:
        node = g.node[module]
        if not node:
            continue
        ns = node["namespace"]
        ns_path = _get_sparse_persistence_file(ns)
        if not os.path.isfile(ns_path):
            all_white_listed.append(module)
        if node["manifest"].get("auto_install"):
            auto_install.append(module)

    for f_path in all_sparse_files:
        with open(f_path, "r") as f:
            all_white_listed.extend(f.read().splitlines())

    state_change = False
    for module in auto_install:
        node = g.node[module]
        if not node:
            continue
        deps = node["manifest"].get("depends", [])
        if not all(dep in all_white_listed for dep in deps):
            continue

        # If no sparse-persistence file exists, no need to whitelist, either.
        ns = node["namespace"]
        ns_path = _get_sparse_persistence_file(ns)
        if not os.path.isfile(ns_path):
            continue

        # If already whitelisted, no need to add it.
        with open(ns_path, "r") as f:
            existing = set(f.read().splitlines())
        if module in existing:
            continue

        # Add it.
        with open(ns_path, "a") as f:
            state_change = True
            f.write(module + "\n")

    return state_change


def ensure_dockerignore_updated(g):
    all_sparse_files = _get_all_sparse_files(g)
    dockerignore_snippet = ""
    for file in all_sparse_files:
        ns = _get_ns_from_sparse_persistence_file(file)
        ignore = [os.path.join(ns, "**")]
        with open(file, "r") as f:
            existing = set(f.read().splitlines())
        less = ["!" + os.path.join(ns, l) for l in existing if "!setup" not in l]
        dockerignore_snippet += "\n".join(ignore + sorted(less)) + "\n"

    with open(".dockerignore", "r") as f:
        lines = f.read().splitlines()

    with open(".dockerignore", "w") as f:
        for line in lines:
            f.write(line + "\n")
            if DOCKERIGNORE_PLACEHOLDER in line:
                break
        f.write(dockerignore_snippet)


def _handle_module(g, module, rootpath, skip_native):
    try:
        deps = ancestors(g, module)
    except networkx.exception.NetworkXError:
        click.secho(
            "UNKNOWN MODULE: '{}' is not in the module graph built from "
            "{}.".format(module, rootpath),
            fg="red",
            bold=True,
        )
        click.get_current_context().exit(code=1)

    node = g.node[module]
    if not node:
        click.secho(
            "MISSING MODULE, BUT REFERENCED: While '{}' is itself listed as a "
            "dependency somewhere, it was found nowhere under "
            "{}.".format(module, rootpath),
            fg="red",
            bold=True,
        )
        click.get_current_context().exit(code=1)

    _warn_path_length(g, deps)

    if skip_native and "vendor/odoo" in node["namespace"]:
        click.get_current_context().exit(
            "You have specified a native module while skipping native modules "
            "from whitelisting."
        )

    # src folder modules should not be white listed
    if "src" in node["namespace"]:
        include = {}
    else:
        include = {node["namespace"]: {module}}

    fail = False
    for dep in deps:
        node = g.node[dep]
        if not node:
            fail = True
            click.secho(
                "MISSING DEPENDENCY: The dependency '{}' was found nowhere "
                "under {}.".format(dep, rootpath),
                fg="red",
                bold=True,
            )
            continue
        ns = node["namespace"]
        if skip_native and "vendor/odoo" in node["namespace"]:
            continue
        # src folder dependencies should not be white listed
        if "src" in node["namespace"]:
            continue
        include.setdefault(ns, set())
        include[ns] |= {dep}
    if fail:
        click.get_current_context().exit(code=1)

    for ns in include.keys():
        ns_path = _get_sparse_persistence_file(ns)
        if os.path.isfile(ns_path):
            with open(ns_path, "r") as f:
                existing = set(f.read().splitlines())
        else:
            existing = set()

        missing = include[ns] - existing
        if not missing:
            continue
        should = include[ns] | existing
        # Make sure setup exepmtion is the last item
        if "!setup/**" in should:
            should.remove("!setup/**")
        with open(ns_path, "w") as f:
            f.write("\n".join(should) + "\n!setup/**\n")


@click.command()
@click.option(
    "--skip-native",
    is_flag=True,
    default=True,
    prompt="Ignore native modules from sparse checkout config?",
    help="Excludes native modules form sparse checkout configuration.",
)
@click.argument("module", required=False)
def whitelist(module, skip_native):
    """ Whitleist a module dependency tree for sparse checkout. If no module is
     specified, whitelist all depedencies of all modules listed in ./src."""

    # We are inside of a git
    if not (
        call_cmd("git rev-parse --is-inside-work-tree", exit_on_error=False) == "true"
    ):
        click.get_current_context().fail("You are not inside a work tree.")

    # Validate we are in the right folder (~/odoo/org/project)
    repo_url = call_cmd("git config --local remote.origin.url")
    if not repo_url:
        click.get_current_context().fail(
            "This project has no origin repo (yet). Please configure an origin "
            "first before continuing with white listing operations."
        )
    top_level = call_cmd("git rev-parse --show-toplevel")
    expected_path = get_fs_target(repo_url)
    if expected_path != top_level:
        click.get_current_context().fail(
            "How you dare not to stick with the odooup folder convention? "
            "This project is expected to live in {}".format(expected_path)
        )

    # Start white listing
    g = get_graph(top_level)
    # If no module is set, whitelist based on src folder
    if not module:
        for module in g:
            node = g.node[module]
            if "namespace" in node and "src/" in node["namespace"]:
                _handle_module(g, module, top_level, skip_native)
    else:
        _handle_module(g, module, top_level, skip_native)

    ensure_sparse_checkouts(top_level)

    while _reconcile_auto_install(g):
        pass

    ensure_dockerignore_updated(g)

    _warn_missing_dependencies(g, top_level)


if __name__ == "__main__":
    whitelist()



import os

import click

from ._cache import cache_repo
from ._helpers import call_cmd, get_fs_target, mkdir_p, parse_git_url
from ._installers import install_tools
from .whitelist import ensure_sparse_checkouts


def _clone(branch, url):
    repo_cache_dir = cache_repo(*parse_git_url(url))
    reference = "--reference {}".format(repo_cache_dir)
    target = get_fs_target(url)
    call_cmd(
        "git clone -b {branch} {reference} --dissociate "
        "{url} {target}".format(**locals()),
        echo_cmd=True,
        exit_on_error=True,
    )
    return target


def _clone_submodules(branch, target, dissociate):
    submodules = [
        s.split(" ")[1]
        for s in call_cmd(
            "git submodule", echo_cmd=False, exit_on_error=True, cwd=target
        ).split("\n")
        if s
    ]
    for submodule in submodules:
        name = call_cmd(
            "git submodule--helper name {}".format(submodule),
            echo_cmd=False,
            exit_on_error=True,
            cwd=target,
        )
        url = call_cmd(
            "git submodule--helper config submodule.{}.url".format(name),
            echo_cmd=False,
            exit_on_error=True,
            cwd=target,
        )
        if url.startswith(".") or url.startswith(".."):
            url = call_cmd(
                "git submodule--helper resolve-relative-url {}".format(url),
                echo_cmd=False,
                exit_on_error=True,
                cwd=target,
            )
        repo_cache_dir = cache_repo(*parse_git_url(url))
        reference = "--reference {}".format(repo_cache_dir)
        dissociate = "--dissociate" if dissociate else ""
        call_cmd(
            "git submodule update {reference} --init {dissociate} "
            "-- {submodule}".format(**locals()),
            echo_cmd=True,
            exit_on_error=False,
            cwd=target,
        )
    call_cmd(
        "git submodule foreach git config remote.origin.fetch "
        "+refs/heads/{branch}:refs/remotes/origin/{branch}".format(**locals()),
        echo_cmd=False,
        exit_on_error=True,
        cwd=target,
    )
    call_cmd(
        "git submodule foreach git fetch --all --prune",
        echo_cmd=False,
        exit_on_error=True,
        cwd=target,
    )


def clone_submodule_to_target(branch, url, target):
    repo_cache_dir = cache_repo(*parse_git_url(url))
    reference = "--reference {}".format(repo_cache_dir)
    call_cmd(
        "git submodule add -b {branch} {reference} --dissociate "
        "{url} {target}".format(**locals()),
        echo_cmd=True,
        exit_on_error=False,
    )


def get_vendor_target(repo_url):
    _, _, org, project = parse_git_url(repo_url)
    mkdir_p(os.path.join("vendor", org))
    return os.path.join("vendor", org, project)


@click.command()
@click.option(
    "--whitelist",
    is_flag=True,
    default=True,
    help="Initialize sparse checkout module whitelisting.",
)
@click.option(
    "--dissociate",
    is_flag=True,
    default=False,
    help="Dissociate cloned submodules from cache.",
)
@click.argument("branch", required=True)
@click.argument("url", required=True)
def clone(branch, url, whitelist, dissociate):
    install_tools()
    target = _clone(branch, url)
    _clone_submodules(branch, target, dissociate)
    if whitelist:
        ensure_sparse_checkouts(target)

    call_cmd("git config commit.template $(pwd)/.git-commit-template")
    call_cmd("pre-commit install --hook-type pre-commit")
    call_cmd("pre-commit install --hook-type commit-msg")
    call_cmd("pre-commit install --install-hooks")


if __name__ == "__main__":
    clone()
